/* extracted from Id */
#define S_NOT_OPEN_SOURCE      1
#define S_NOT_OPEN_TARGET      2
#define S_NOT_OPEN_FILE        3
#define S_NOT_STAT_FILE        4
#define S_SOURCE_READ_ERR      5
#define S_READ_ERROR           6
#define S_IO_ERR_TARGET        7
#define S_WR_ERR_TARGET        8
#define S_WR_ERR_TEMP          9
#define S_TMP_NOT_REMOVED     10
#define S_OUT_OF_MEMORY       11
#define S_TARGET_EXISTS       12
#define S_NOT_RENAME          13
#define S_ERR_ENCODING        14
#define S_STAT_ONE_PART       15
#define S_PARM_CHECK          16
#define S_SHORT_BINHEX        17
#define S_DECODE_CANCEL       18
#define S_ENCODE_CANCEL       19
#define S_SCAN_CANCEL         20
#define S_SIZE_MISMATCH       21
#define S_PSIZE_MISMATCH      22
#define S_CRC_MISMATCH        23
#define S_PCRC_MISMATCH       24
#define S_LOADED_PART         25
#define S_NO_DATA_FOUND       26
#define S_NO_BIN_FILE         27
#define S_STRIPPED_SETUID     28
#define S_DATA_SUSPICIOUS     29
#define S_NO_TEMP_NAME        30
#define S_BINHEX_SIZES        31
#define S_BINHEX_BOTH         32
#define S_SMERGE_MERGED       33
#define S_MIME_NO_BOUNDARY    34
#define S_MIME_B_NOT_FOUND    35
#define S_MIME_MULTI_DEPTH    36
#define S_MIME_PART_MULTI     37
